package tart.test

import android.app.AlertDialog
import android.os.SystemClock
import android.view.MotionEvent
import android.view.View
import androidx.test.core.app.ActivityScenario
import androidx.test.espresso.Espresso.onView
import androidx.test.espresso.UiController
import androidx.test.espresso.ViewAction
import androidx.test.espresso.matcher.RootMatchers.isDialog
import androidx.test.espresso.matcher.ViewMatchers
import androidx.test.espresso.matcher.ViewMatchers.withId
import androidx.test.platform.app.InstrumentationRegistry
import com.google.common.truth.Truth.assertThat
import com.google.common.truth.Truth.assertWithMessage
import org.hamcrest.Matcher
import org.junit.Test
import tart.internal.FrozenFrameOnTouchDetector
import tart.legacy.ActivityEvent
import tart.legacy.AppStart.AppStartData
import tart.legacy.FrozenFrameOnTouch
import tart.legacy.Perfs
import tart.test.utilities.TestActivity
import java.util.concurrent.CountDownLatch
import java.util.concurrent.TimeUnit.SECONDS
import java.util.concurrent.atomic.AtomicReference

class PerfMonitoringTest {

  /**
   * By design [Perfs] is hooked deeply into Android internals and is made to execute and init
   * exactly once per process, whereas we typically run several instrumentation tests per process.
   * Unless or until we use Orchestrator, we have to write these tests as a series of steps that
   * build on top of each other.
   */
  @Test
  fun allTestExecutingSeriallyInOneProcess() {
    initialAppStart()
    firstActivity()
    customEvents()
    frozenFrames()
  }

  private fun initialAppStart() {
    assertThat(appStart.processStartUptimeMillis)
      .isLessThan(SystemClock.uptimeMillis())
    assertThat(appStart.firstAppClassLoadElapsedUptimeMillis).isGreaterThan(0)
    assertThat(appStart.firstComponentInstantiated).isNull()
  }

  private fun firstActivity() {
    ActivityScenario.launch(TestActivity::class.java).use {
      // Our activity isn't actually first when testing, oh well.
      val firstActivity = "androidx.test.core.app.InstrumentationActivityInvoker\$BootstrapActivity"
      assertThat(appStart.firstActivityOnCreate!!.activityName).isEqualTo(
        firstActivity
      )

      val firstActivityOnCreate =
        appStart.firstActivityOnCreate?.let {
          ActivityEvent(
            it.activityName,
            it.elapsedUptimeMillis
          )
        }

      var previousElapsed = appStart.firstAppClassLoadElapsedUptimeMillis
      for (
      (activityEvent, lifecycle) in listOf(
        firstActivityOnCreate to "onCreate()",
        appStart.firstActivityOnStart to "onStart()",
        appStart.firstActivityOnResume to "onResume()",
        // Unfortunately the bootstrap activity doesn't get laid out / drawn
        // appStart.firstGlobalLayout to "onGlobalLayout()",
        // appStart.firstPreDraw to "onPreDraw()",
        // appStart.firstDraw to "onDraw()"
      )
      ) {
        assertWithMessage("For $lifecycle").that(activityEvent)
          .isNotNull()
        assertWithMessage("For $lifecycle").that(activityEvent!!.activityName)
          .isEqualTo(firstActivity)
        assertWithMessage("For $lifecycle").that(activityEvent.elapsedUptimeMillis)
          .isAtLeast(previousElapsed)
        previousElapsed = activityEvent.elapsedUptimeMillis
      }
    }
  }

  private fun customEvents() {
    runOnMainSync {
      Perfs.customFirstEvent(eventName = "Kouign-amann")
      SystemClock.sleep(500)
      Perfs.customFirstEvent(eventName = "Croissant", extra = "Pain au chocolat")
      Perfs.customFirstEvent(eventName = "Croissant", extra = "Chocolatine")
    }

    assertThat(appStart.customFirstEvents["Kouign-amann"]?.first).isLessThan(
      appStart.customFirstEvents["Croissant"]?.first
    )
    assertThat(appStart.customFirstEvents["Croissant"]?.second).isEqualTo("Pain au chocolat")
  }

  private fun frozenFrames() {
    val waitForFrozenFrame = reportFrozenFrame()

    ActivityScenario.launch(TestActivity::class.java).use { scenario ->

      scenario.onActivity { activity ->
        AlertDialog.Builder(activity)
          // The tap is late (late, latte) and it leads to frozen frames (frozen, frappé coffee)
          .setPositiveButton("Better latte than frappé", null)
          .show()
      }

      val (buttonX, buttonY) = findDialogButtonCoordinates(android.R.id.button1)

      sendTapAtTime(SystemClock.uptimeMillis() - 2000, buttonX, buttonY)

      val frozenFrameOnTouch = waitForFrozenFrame()

      assertThat(frozenFrameOnTouch.pressedView).contains("android:id/button1")
      assertThat(frozenFrameOnTouch.handledElapsedUptimeMillis).isAtLeast(2000)
    }
  }

  private fun reportFrozenFrame(): () -> FrozenFrameOnTouch {
    val waitForFrozenFrame = CountDownLatch(1)
    val frozenFrameOnTouchRef = AtomicReference<FrozenFrameOnTouch>()
    runOnMainSync {
      FrozenFrameOnTouchDetector.install { frozenFrameOnTouch ->
        frozenFrameOnTouchRef.set(frozenFrameOnTouch)
        waitForFrozenFrame.countDown()
      }
    }
    return {
      require(waitForFrozenFrame.await(10, SECONDS))
      frozenFrameOnTouchRef.get()!!
    }
  }

  private fun runOnMainSync(block: () -> Unit) {
    InstrumentationRegistry.getInstrumentation().runOnMainSync(block)
  }

  private fun findDialogButtonCoordinates(buttonResId: Int): Pair<Int, Int> {
    val waitForDialogButton = CountDownLatch(1)
    val dialogButtonCenter = AtomicReference<Pair<Int, Int>>()
    onView(withId(buttonResId)).inRoot(isDialog())
      .perform(object : ViewAction {
        override fun getDescription() = "Retrieving dialog ok button"

        override fun getConstraints(): Matcher<View> = ViewMatchers.isDisplayed()

        override fun perform(
          uiController: UiController,
          view: View
        ) {
          val location = IntArray(2)
          view.getLocationOnScreen(location)
          val centerX = location[0] + view.width / 2
          val centerY = location[1] + view.height / 2
          dialogButtonCenter.set(centerX to centerY)
          waitForDialogButton.countDown()
        }
      })
    require(waitForDialogButton.await(10, SECONDS))
    return dialogButtonCenter.get()!!
  }

  private fun sendTapAtTime(
    downTime: Long,
    clickX: Int,
    clickY: Int
  ) {
    val instrumentation = InstrumentationRegistry.getInstrumentation()
    instrumentation.sendPointerSync(
      MotionEvent.obtain(
        downTime,
        downTime,
        MotionEvent.ACTION_DOWN,
        clickX.toFloat(),
        clickY.toFloat(),
        0
      )
    )
    instrumentation.sendPointerSync(
      MotionEvent.obtain(
        downTime,
        downTime + 50,
        MotionEvent.ACTION_UP,
        clickX.toFloat(),
        clickY.toFloat(),
        0
      )
    )
  }

  private val appStart: AppStartData
    get() = Perfs.appStart as AppStartData
}